"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.QuickFixXmlFileBuilder = void 0;
const buffer_1 = require("../../../buffer");
const contained_1 = require("../../contained");
const lodash_1 = require("lodash");
const collections_1 = require("../../../collections");
const quick_fix_xml_formatter_1 = require("./quick-fix-xml-formatter");
class QuickFixXmlFileBuilder {
    constructor(definitions) {
        this.definitions = definitions;
        this.usedTags = {};
        this.requiredComponents = [];
        this.seenComponents = new collections_1.Dictionary();
        this.indent = 2;
        this.dispatcher = new contained_1.FieldsDispatch();
        this.elasticBuffer = new buffer_1.ElasticBuffer(10 * 1024);
    }
    write(msgTypes) {
        const eb = this.elasticBuffer;
        if (msgTypes == null || msgTypes.length === 0)
            return;
        eb.reset();
        const m0def = this.definitions.message.get(msgTypes[0]);
        eb.writeString(quick_fix_xml_formatter_1.QuickFixXmlFormatter.startFix(this.definitions.getMajor(), this.definitions.getMinor(), this.definitions.getServicePack()));
        const header = this.writeComponent(`${m0def === null || m0def === void 0 ? void 0 : m0def.name}.StandardHeader`, 'header');
        eb.writeString(header);
        const trailer = this.writeComponent(`${m0def === null || m0def === void 0 ? void 0 : m0def.name}.StandardTrailer`, 'trailer');
        eb.writeString(trailer);
        const messages = this.writeMessages(msgTypes, this.indent);
        eb.writeString(messages);
        const components = this.writeComponents(this.indent);
        eb.writeString(components);
        const fields = this.writeFieldDefinitions(this.indent);
        eb.writeString(fields);
        eb.writeString(quick_fix_xml_formatter_1.QuickFixXmlFormatter.endFix());
    }
    writeComponents(leadingIndent) {
        const eb = new buffer_1.ElasticBuffer(2 * 1024);
        eb.writeString(quick_fix_xml_formatter_1.QuickFixXmlFormatter.startEntity('components', leadingIndent));
        const components = this.requiredComponents;
        while (components.length > 0) {
            const k = components.pop();
            if (!k)
                continue;
            const component = this.definitions.component.get(k);
            if (!component)
                continue;
            eb.writeString(quick_fix_xml_formatter_1.QuickFixXmlFormatter.startComponent(component.name, leadingIndent + this.indent));
            const def = this.writeFields(component.fields, leadingIndent + this.indent + this.indent);
            eb.writeString(def);
            eb.writeString(quick_fix_xml_formatter_1.QuickFixXmlFormatter.endComponent(leadingIndent + this.indent));
        }
        eb.writeString(quick_fix_xml_formatter_1.QuickFixXmlFormatter.endEntity('components', leadingIndent));
        return eb.toString();
    }
    writeMessages(msgTypes, leadingIndent) {
        const eb = new buffer_1.ElasticBuffer(2 * 1024);
        eb.writeString(quick_fix_xml_formatter_1.QuickFixXmlFormatter.startEntity('messages', leadingIndent));
        msgTypes.forEach(mt => {
            const md = this.definitions.message.get(mt);
            if (!md)
                return;
            eb.writeString(quick_fix_xml_formatter_1.QuickFixXmlFormatter.defineMessage(md, leadingIndent + this.indent));
            const fields = this.writeFields(md.fields.slice(1, md.fields.length - 1), leadingIndent + this.indent + this.indent);
            eb.writeString(fields);
            eb.writeString(quick_fix_xml_formatter_1.QuickFixXmlFormatter.endEntity('message', leadingIndent + this.indent));
        });
        eb.writeString(quick_fix_xml_formatter_1.QuickFixXmlFormatter.endEntity('messages', leadingIndent));
        return eb.toString();
    }
    writeFieldDefinitions(leadingIndent) {
        const eb = new buffer_1.ElasticBuffer(2 * 1024);
        const tags = (0, lodash_1.keys)(this.usedTags).map(s => parseInt(s));
        tags.sort(function (a, b) {
            return a - b;
        });
        eb.writeString(quick_fix_xml_formatter_1.QuickFixXmlFormatter.startEntity('fields', leadingIndent));
        tags.forEach(t => {
            const sf = this.definitions.tagToSimple[t];
            if (!sf)
                return;
            eb.writeString(quick_fix_xml_formatter_1.QuickFixXmlFormatter.defineField(sf, leadingIndent + this.indent));
            if (sf.isEnum()) {
                const en = this.writeEnumDefinition(sf, leadingIndent + this.indent + this.indent);
                eb.writeString(en);
                eb.writeString(quick_fix_xml_formatter_1.QuickFixXmlFormatter.endEntity('field', leadingIndent + this.indent));
            }
        });
        eb.writeString(quick_fix_xml_formatter_1.QuickFixXmlFormatter.endEntity('fields', leadingIndent));
        return eb.toString();
    }
    writeEnumDefinition(sf, leadingIndent) {
        const eb = new buffer_1.ElasticBuffer(2 * 1024);
        const keys = sf.enums.keys();
        keys.sort();
        keys.forEach(k => {
            eb.writeString(quick_fix_xml_formatter_1.QuickFixXmlFormatter.addEnum(sf.enums.get(k), leadingIndent));
        });
        return eb.toString();
    }
    writeComponent(name, destName) {
        const eb = new buffer_1.ElasticBuffer(2 * 1024);
        const set = this.definitions.getSet(name);
        if (set == null)
            return '';
        eb.writeString(quick_fix_xml_formatter_1.QuickFixXmlFormatter.startEntity(destName, this.indent));
        eb.writeString(this.writeFields(set.fields, this.indent + this.indent));
        eb.writeString(quick_fix_xml_formatter_1.QuickFixXmlFormatter.endEntity(destName, this.indent));
        return eb.toString();
    }
    writeSimpleField(sf, eb, leadingIndent) {
        eb.writeString(quick_fix_xml_formatter_1.QuickFixXmlFormatter.addField(sf, leadingIndent));
        this.usedTags[sf.definition.tag] = sf.name;
    }
    writeComponentField(cf, eb, leadingIndent) {
        eb.writeString(quick_fix_xml_formatter_1.QuickFixXmlFormatter.addComponent(cf, leadingIndent));
        if (!this.seenComponents.containsKey(cf.name)) {
            this.seenComponents.add(cf.name, true);
            this.requiredComponents.push(cf.name);
        }
    }
    writeGroupField(gf, eb, leadingIndent) {
        var _a, _b;
        this.usedTags[(_b = (_a = gf.definition.noOfField) === null || _a === void 0 ? void 0 : _a.tag) !== null && _b !== void 0 ? _b : 0] = gf.name;
        eb.writeString(quick_fix_xml_formatter_1.QuickFixXmlFormatter.addGroup(gf, leadingIndent));
        const groupDef = this.writeFields(gf.definition.fields, leadingIndent + this.indent);
        eb.writeString(groupDef);
        eb.writeString(quick_fix_xml_formatter_1.QuickFixXmlFormatter.endGroup(leadingIndent));
    }
    writeFields(fields, leadingIndent) {
        const eb = new buffer_1.ElasticBuffer(2 * 1024);
        this.dispatcher.dispatchFields(fields, {
            simple: (sf) => { this.writeSimpleField(sf, eb, leadingIndent); },
            component: (cf) => { this.writeComponentField(cf, eb, leadingIndent); },
            group: (gf) => { this.writeGroupField(gf, eb, leadingIndent); }
        });
        return eb.toString();
    }
}
exports.QuickFixXmlFileBuilder = QuickFixXmlFileBuilder;
//# sourceMappingURL=quick-fix-xml-file-builder.js.map